'use strict';

var domain;

// This constructor is used to store event handlers. Instantiating this is
// faster than explicitly calling `Object.create(null)` to get a "clean" empty
// object (tested with v8 v4.9).
function EventHandlers() {}
EventHandlers.prototype = Object.create(null);

function EventEmitter() {
  EventEmitter.init.call(this);
}

// nodejs oddity
// require('events') === require('events').EventEmitter
EventEmitter.EventEmitter = EventEmitter;

EventEmitter.usingDomains = false;

EventEmitter.prototype.domain = undefined;
EventEmitter.prototype._events = undefined;
EventEmitter.prototype._maxListeners = undefined;

// By default EventEmitters will print a warning if more than 10 listeners are
// added to it. This is a useful default which helps finding memory leaks.
EventEmitter.defaultMaxListeners = 10;

EventEmitter.init = function() {
  this.domain = null;
  if (EventEmitter.usingDomains) {
    // if there is an active domain, then attach to it.
    if (domain.active ) ;
  }

  if (!this._events || this._events === Object.getPrototypeOf(this)._events) {
    this._events = new EventHandlers();
    this._eventsCount = 0;
  }

  this._maxListeners = this._maxListeners || undefined;
};

// Obviously not all Emitters should be limited to 10. This function allows
// that to be increased. Set to zero for unlimited.
EventEmitter.prototype.setMaxListeners = function setMaxListeners(n) {
  if (typeof n !== 'number' || n < 0 || isNaN(n))
    throw new TypeError('"n" argument must be a positive number');
  this._maxListeners = n;
  return this;
};

function $getMaxListeners(that) {
  if (that._maxListeners === undefined)
    return EventEmitter.defaultMaxListeners;
  return that._maxListeners;
}

EventEmitter.prototype.getMaxListeners = function getMaxListeners() {
  return $getMaxListeners(this);
};

// These standalone emit* functions are used to optimize calling of event
// handlers for fast cases because emit() itself often has a variable number of
// arguments and can be deoptimized because of that. These functions always have
// the same number of arguments and thus do not get deoptimized, so the code
// inside them can execute faster.
function emitNone(handler, isFn, self) {
  if (isFn)
    handler.call(self);
  else {
    var len = handler.length;
    var listeners = arrayClone(handler, len);
    for (var i = 0; i < len; ++i)
      listeners[i].call(self);
  }
}
function emitOne(handler, isFn, self, arg1) {
  if (isFn)
    handler.call(self, arg1);
  else {
    var len = handler.length;
    var listeners = arrayClone(handler, len);
    for (var i = 0; i < len; ++i)
      listeners[i].call(self, arg1);
  }
}
function emitTwo(handler, isFn, self, arg1, arg2) {
  if (isFn)
    handler.call(self, arg1, arg2);
  else {
    var len = handler.length;
    var listeners = arrayClone(handler, len);
    for (var i = 0; i < len; ++i)
      listeners[i].call(self, arg1, arg2);
  }
}
function emitThree(handler, isFn, self, arg1, arg2, arg3) {
  if (isFn)
    handler.call(self, arg1, arg2, arg3);
  else {
    var len = handler.length;
    var listeners = arrayClone(handler, len);
    for (var i = 0; i < len; ++i)
      listeners[i].call(self, arg1, arg2, arg3);
  }
}

function emitMany(handler, isFn, self, args) {
  if (isFn)
    handler.apply(self, args);
  else {
    var len = handler.length;
    var listeners = arrayClone(handler, len);
    for (var i = 0; i < len; ++i)
      listeners[i].apply(self, args);
  }
}

EventEmitter.prototype.emit = function emit(type) {
  var er, handler, len, args, i, events, domain;
  var doError = (type === 'error');

  events = this._events;
  if (events)
    doError = (doError && events.error == null);
  else if (!doError)
    return false;

  domain = this.domain;

  // If there is no 'error' event listener then throw.
  if (doError) {
    er = arguments[1];
    if (domain) {
      if (!er)
        er = new Error('Uncaught, unspecified "error" event');
      er.domainEmitter = this;
      er.domain = domain;
      er.domainThrown = false;
      domain.emit('error', er);
    } else if (er instanceof Error) {
      throw er; // Unhandled 'error' event
    } else {
      // At least give some kind of context to the user
      var err = new Error('Uncaught, unspecified "error" event. (' + er + ')');
      err.context = er;
      throw err;
    }
    return false;
  }

  handler = events[type];

  if (!handler)
    return false;

  var isFn = typeof handler === 'function';
  len = arguments.length;
  switch (len) {
    // fast cases
    case 1:
      emitNone(handler, isFn, this);
      break;
    case 2:
      emitOne(handler, isFn, this, arguments[1]);
      break;
    case 3:
      emitTwo(handler, isFn, this, arguments[1], arguments[2]);
      break;
    case 4:
      emitThree(handler, isFn, this, arguments[1], arguments[2], arguments[3]);
      break;
    // slower
    default:
      args = new Array(len - 1);
      for (i = 1; i < len; i++)
        args[i - 1] = arguments[i];
      emitMany(handler, isFn, this, args);
  }

  return true;
};

function _addListener(target, type, listener, prepend) {
  var m;
  var events;
  var existing;

  if (typeof listener !== 'function')
    throw new TypeError('"listener" argument must be a function');

  events = target._events;
  if (!events) {
    events = target._events = new EventHandlers();
    target._eventsCount = 0;
  } else {
    // To avoid recursion in the case that type === "newListener"! Before
    // adding it to the listeners, first emit "newListener".
    if (events.newListener) {
      target.emit('newListener', type,
                  listener.listener ? listener.listener : listener);

      // Re-assign `events` because a newListener handler could have caused the
      // this._events to be assigned to a new object
      events = target._events;
    }
    existing = events[type];
  }

  if (!existing) {
    // Optimize the case of one listener. Don't need the extra array object.
    existing = events[type] = listener;
    ++target._eventsCount;
  } else {
    if (typeof existing === 'function') {
      // Adding the second element, need to change to array.
      existing = events[type] = prepend ? [listener, existing] :
                                          [existing, listener];
    } else {
      // If we've already got an array, just append.
      if (prepend) {
        existing.unshift(listener);
      } else {
        existing.push(listener);
      }
    }

    // Check for listener leak
    if (!existing.warned) {
      m = $getMaxListeners(target);
      if (m && m > 0 && existing.length > m) {
        existing.warned = true;
        var w = new Error('Possible EventEmitter memory leak detected. ' +
                            existing.length + ' ' + type + ' listeners added. ' +
                            'Use emitter.setMaxListeners() to increase limit');
        w.name = 'MaxListenersExceededWarning';
        w.emitter = target;
        w.type = type;
        w.count = existing.length;
        emitWarning(w);
      }
    }
  }

  return target;
}
function emitWarning(e) {
  typeof console.warn === 'function' ? console.warn(e) : console.log(e);
}
EventEmitter.prototype.addListener = function addListener(type, listener) {
  return _addListener(this, type, listener, false);
};

EventEmitter.prototype.on = EventEmitter.prototype.addListener;

EventEmitter.prototype.prependListener =
    function prependListener(type, listener) {
      return _addListener(this, type, listener, true);
    };

function _onceWrap(target, type, listener) {
  var fired = false;
  function g() {
    target.removeListener(type, g);
    if (!fired) {
      fired = true;
      listener.apply(target, arguments);
    }
  }
  g.listener = listener;
  return g;
}

EventEmitter.prototype.once = function once(type, listener) {
  if (typeof listener !== 'function')
    throw new TypeError('"listener" argument must be a function');
  this.on(type, _onceWrap(this, type, listener));
  return this;
};

EventEmitter.prototype.prependOnceListener =
    function prependOnceListener(type, listener) {
      if (typeof listener !== 'function')
        throw new TypeError('"listener" argument must be a function');
      this.prependListener(type, _onceWrap(this, type, listener));
      return this;
    };

// emits a 'removeListener' event iff the listener was removed
EventEmitter.prototype.removeListener =
    function removeListener(type, listener) {
      var list, events, position, i, originalListener;

      if (typeof listener !== 'function')
        throw new TypeError('"listener" argument must be a function');

      events = this._events;
      if (!events)
        return this;

      list = events[type];
      if (!list)
        return this;

      if (list === listener || (list.listener && list.listener === listener)) {
        if (--this._eventsCount === 0)
          this._events = new EventHandlers();
        else {
          delete events[type];
          if (events.removeListener)
            this.emit('removeListener', type, list.listener || listener);
        }
      } else if (typeof list !== 'function') {
        position = -1;

        for (i = list.length; i-- > 0;) {
          if (list[i] === listener ||
              (list[i].listener && list[i].listener === listener)) {
            originalListener = list[i].listener;
            position = i;
            break;
          }
        }

        if (position < 0)
          return this;

        if (list.length === 1) {
          list[0] = undefined;
          if (--this._eventsCount === 0) {
            this._events = new EventHandlers();
            return this;
          } else {
            delete events[type];
          }
        } else {
          spliceOne(list, position);
        }

        if (events.removeListener)
          this.emit('removeListener', type, originalListener || listener);
      }

      return this;
    };
    
// Alias for removeListener added in NodeJS 10.0
// https://nodejs.org/api/events.html#events_emitter_off_eventname_listener
EventEmitter.prototype.off = function(type, listener){
    return this.removeListener(type, listener);
};

EventEmitter.prototype.removeAllListeners =
    function removeAllListeners(type) {
      var listeners, events;

      events = this._events;
      if (!events)
        return this;

      // not listening for removeListener, no need to emit
      if (!events.removeListener) {
        if (arguments.length === 0) {
          this._events = new EventHandlers();
          this._eventsCount = 0;
        } else if (events[type]) {
          if (--this._eventsCount === 0)
            this._events = new EventHandlers();
          else
            delete events[type];
        }
        return this;
      }

      // emit removeListener for all listeners on all events
      if (arguments.length === 0) {
        var keys = Object.keys(events);
        for (var i = 0, key; i < keys.length; ++i) {
          key = keys[i];
          if (key === 'removeListener') continue;
          this.removeAllListeners(key);
        }
        this.removeAllListeners('removeListener');
        this._events = new EventHandlers();
        this._eventsCount = 0;
        return this;
      }

      listeners = events[type];

      if (typeof listeners === 'function') {
        this.removeListener(type, listeners);
      } else if (listeners) {
        // LIFO order
        do {
          this.removeListener(type, listeners[listeners.length - 1]);
        } while (listeners[0]);
      }

      return this;
    };

EventEmitter.prototype.listeners = function listeners(type) {
  var evlistener;
  var ret;
  var events = this._events;

  if (!events)
    ret = [];
  else {
    evlistener = events[type];
    if (!evlistener)
      ret = [];
    else if (typeof evlistener === 'function')
      ret = [evlistener.listener || evlistener];
    else
      ret = unwrapListeners(evlistener);
  }

  return ret;
};

EventEmitter.listenerCount = function(emitter, type) {
  if (typeof emitter.listenerCount === 'function') {
    return emitter.listenerCount(type);
  } else {
    return listenerCount.call(emitter, type);
  }
};

EventEmitter.prototype.listenerCount = listenerCount;
function listenerCount(type) {
  var events = this._events;

  if (events) {
    var evlistener = events[type];

    if (typeof evlistener === 'function') {
      return 1;
    } else if (evlistener) {
      return evlistener.length;
    }
  }

  return 0;
}

EventEmitter.prototype.eventNames = function eventNames() {
  return this._eventsCount > 0 ? Reflect.ownKeys(this._events) : [];
};

// About 1.5x faster than the two-arg version of Array#splice().
function spliceOne(list, index) {
  for (var i = index, k = i + 1, n = list.length; k < n; i += 1, k += 1)
    list[i] = list[k];
  list.pop();
}

function arrayClone(arr, i) {
  var copy = new Array(i);
  while (i--)
    copy[i] = arr[i];
  return copy;
}

function unwrapListeners(arr) {
  var ret = new Array(arr.length);
  for (var i = 0; i < ret.length; ++i) {
    ret[i] = arr[i].listener || arr[i];
  }
  return ret;
}

function cleanQuery(object) {
  Object.keys(object).forEach(key => !object[key] && delete object[key]);
  return object;
}
function cleanNullsorUndefined(object) {
  Object.keys(object).forEach(key => (object[key] === null || object[key] === undefined) && delete object[key]);
  return object;
}

/* eslint camelcase: "off" */

const Training = function (fetch, baseUrl, token) {
  const api = {
    activityLogs: {
      list(trainingId, filters = {}, options = {}) {
        const query = {
          training: trainingId,
          page: options.page
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v1/activity-logs/?${qs}`);
      },

      get(id) {
        return api.fetch(`v1/activity-logs/${id}/`);
      }

    },
    languages: {
      list() {
        return api.fetch(`v1/languages/`);
      },

      listAll() {
        return api.fetch(`v1/languages/all/`);
      }

    },
    translationPackages: {
      list({
        projectId,
        assignee,
        state
      } = {}) {
        const query = {
          project: projectId,
          assignee,
          state
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v3/translation-packages/?${qs}`);
      },

      get(id) {
        return api.fetch(`v3/translation-packages/${id}/`);
      },

      create(translationPackage) {
        return api.fetch(`v3/translation-packages/`, 'POST', translationPackage);
      },

      delete(id) {
        return api.fetch(`v3/translation-packages/${id}/`, 'DELETE');
      },

      update(translationPackageId, update) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/`, 'PATCH', update);
      },

      simulate(packageOptions) {
        return api.fetch('v3/translation-packages/simulate/', 'POST', packageOptions);
      }

    },
    translationItems: {
      list(translationPackageId) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/`);
      },

      create({
        translationPackageId,
        translationItem
      }) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/`, 'POST', translationItem);
      },

      delete({
        translationPackageId,
        objectType,
        objectId
      }) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/${objectType}-${objectId}/`, 'DELETE');
      },

      update({
        translationPackageId,
        objectType,
        objectId,
        ...update
      }) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/${objectType}-${objectId}/`, 'PATCH', update);
      }

    },
    notifications: {
      list() {
        return api.fetch(`v3/notifications/`);
      },

      get(id) {
        return api.fetch(`v3/notifications/${id}/`);
      },

      update(id, patchset) {
        return api.fetch(`v3/notifications/${id}/`, 'PATCH', patchset);
      },

      getUnreadCount() {
        return api.fetch('v3/notifications/unread-count/');
      },

      markAllAsRead() {
        return api.fetch('v3/notifications/mark-all-as-read/', 'POST');
      }

    },
    trainings: {
      list(fields = 'id,name') {
        return api.fetch(`v1/trainings/?fields=${fields}`);
      },

      get(id) {
        return api.fetch(`v1/trainings/${id}/`);
      },

      properties(id) {
        return api.fetch(`v1/properties/?training=${id}`);
      },

      create(name, license_holder, more) {
        return api.fetch(`v1/trainings/`, 'POST', {
          name,
          license_holder,
          ...more
        });
      },

      update(id, patchset) {
        return api.fetch(`v1/trainings/${id}/`, 'PATCH', patchset);
      },

      delete(id) {
        return api.fetch(`v1/trainings/${id}/`, 'DELETE');
      },

      publish(id) {
        return api.fetch(`v1/trainings/${id}/publish/`, 'POST');
      }

    },
    // https://github.com/aexeagmbh/rincewind/issues/317#issuecomment-286102626
    validationData: {
      list(trainingId) {
        return api.fetch(`v1/validation-data/?training=${trainingId}`);
      },

      get(id) {
        return api.fetch(`v1/validation-data/${id}/`);
      },

      create(validationData) {
        // {training, name, uid, data}
        return api.fetch(`v1/validation-data/`, 'POST', validationData);
      },

      createSyncedDocument(trainingId, collection_id, document) {
        return api.fetch(`v1/validation-data/create-synced/`, 'POST', {
          training: trainingId,
          name: document.name,
          uid: document.uid,
          data: document.blob,
          myax_api_version: 2,
          myax_reference: {
            collection_id,
            document_id: document.id
          }
        });
      },

      update(validationData) {
        const {
          id,
          name,
          uid,
          data
        } = validationData;
        return api.fetch(`v1/validation-data/${id}/`, 'PATCH', {
          name,
          uid,
          data
        });
      },

      delete(id) {
        return api.fetch(`v1/validation-data/${id}/`, 'DELETE');
      },

      select(id, language) {
        return api.fetch(`v1/validation-data/${id}/select/`, 'PUT', {
          language
        });
      },

      selectDefault(id, language) {
        return api.fetch(`v1/validation-data/${id}/select-global/`, 'PUT', {
          language
        });
      },

      sync(validationData) {
        const {
          id,
          name,
          uid,
          data
        } = validationData;
        return api.fetch(`v1/validation-data/${id}/sync/`, 'PUT', {
          name,
          uid,
          data
        });
      }

    },
    import: {
      import(trainingId, atml) {
        return api.fetch(`v1/trainings/${trainingId}/import/`, 'PUT', {
          atml
        });
      },

      migrate(oldTrainingId, newTrainingId, options) {
        return api.fetch(`v1/trainings/${newTrainingId}/migrate-from-rincewind/`, 'POST', {
          rincewind_training_id: oldTrainingId,
          copy_collections: options && options.copyCollections,
          parametrize_expressions: options && options.parametrizeExpressions
        });
      },

      status(trainingId) {
        return api.fetch(`v1/trainings/${trainingId}/import/?latest=1`);
      }

    }
  };
  api.fetch = fetch;
  return api;
};

/* eslint camelcase: "off" */

const fieldsFromOptions = function (options) {
  return options.fields instanceof Array ? options.fields.join(',') : options.fields;
};

const MyAx = function (fetch, baseUrl, token) {
  const api = {
    instantGenerationEndpoints: {
      list(training_id, options = {}) {
        const query = {
          training_id,
          page_size: options.pageSize || 50,
          fields: fieldsFromOptions(options),
          is_instant: true
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/collections/?${qs}`);
      },

      get(id, options = {}) {
        return api.collections.get(id, options);
      },

      create(endpoint) {
        endpoint.is_instant = true;
        return api.fetch(`v2/collections/`, 'POST', endpoint);
      }

    },
    collections: {
      // options = {page_size: 50, fields}
      list(training_id, options = {}) {
        const query = {
          training_id,
          page_size: options.pageSize || 50,
          fields: fieldsFromOptions(options),
          is_instant: false
        };
        const qs = new URLSearchParams(cleanNullsorUndefined(query)).toString();
        return api.fetch(`v2/collections/?${qs}`);
      },

      get(id, options = {}) {
        const query = {
          fields: fieldsFromOptions(options)
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/collections/${id}/?${qs}`);
      },

      getProcessingStates(collectionId) {
        return api.fetch(`v2/collections/${collectionId}/document-processing-states/`);
      },

      // collection = { language, training_id, name, licence_holder, training_tag: 'draft' }
      create(collection) {
        return api.fetch(`v2/collections/`, 'POST', collection);
      },

      delete(id) {
        return api.fetch(`v2/collections/${id}/`, 'DELETE');
      },

      save(collection) {
        return api.fetch(`v2/collections/${collection.id}/`, 'PATCH', {
          name: collection.name,
          language: collection.language,
          uses_published_training: collection.uses_published_training,
          webhook_secret: collection.webhook_secret,
          webhook_url: collection.webhook_url,
          fail_webhook_url: collection.fail_webhook_url,
          flags: collection.flags
        });
      },

      generateAll(id) {
        return api.fetch(`v2/collections/${id}/generate-content/`, 'POST');
      },

      generateFiltered(id, search, processingState) {
        const query = {
          search,
          processing_state: processingState
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/collections/${id}/generate-content/?${qs}`, 'POST');
      },

      empty(id) {
        return api.fetch(`v2/collections/${id}/empty/`, 'POST');
      }

    },
    // https://github.com/aexeagmbh/myax/blob/master/myax/document/processing_states.py
    // optional search and processingState ['none', 'requested', 'generated', 'delivered', 'failed']
    documents: {
      // filters = {search, processingState}
      // options = {fields}
      list(collectionId, filters = {}, options = {}) {
        const query = {
          collection: collectionId,
          search: filters.search,
          processing_state: filters.processingState,
          fields: options.fields
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/documents/?${qs}`);
      },

      // blob has to contain 'name' and 'uid'
      create(collectionId, blob) {
        return api.fetch(`v2/collections/${collectionId}/document/`, 'POST', blob);
      },

      // blob has to contain 'name'
      updateOrCreateByUid(collectionId, blob, uid) {
        const payload = Object.assign({
          uid
        }, blob);
        return api.fetch(`v2/collections/${collectionId}/update-or-create-by-uid/?uid=${uid}`, 'POST', payload);
      },

      // options = {fields}
      get(id, options = {}) {
        const query = {
          fields: fieldsFromOptions(options)
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/documents/${id}/?${qs}`);
      },

      generate(id) {
        return api.fetch(`v2/documents/${id}/generate-content/?highprio`, 'POST');
      },

      delete(documentId) {
        return api.fetch(`v2/documents/${documentId}/`, 'DELETE');
      }

    },
    histograms: {
      list(collectionOrProjectId, filters = {
        pageSize: 100
      }) {
        const query = {};

        if (typeof collectionOrProjectId === 'number' || /^[0-9]+$/.test(collectionOrProjectId)) {
          query.collection = collectionOrProjectId;
        } else if (collectionOrProjectId) {
          query.project = collectionOrProjectId;
        }

        if (filters && filters.latest) query.latest = true;
        if (filters && filters.pageSize) query.page_size = filters.pageSize;
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/histograms/?${qs}`);
      },

      create(collectionIdOrData) {
        let data;

        if (typeof collectionIdOrData === 'object') {
          data = collectionIdOrData;
        } else {
          data = {
            collection: collectionIdOrData
          };
        }

        return api.fetch(`v2/histograms/`, 'POST', data);
      },

      delete(id) {
        return api.fetch(`v2/histograms/${id}/`, 'DELETE');
      },

      get(id) {
        return api.fetch(`v2/histograms/${id}/`);
      },

      update(id, update) {
        return api.fetch(`v2/histograms/${id}/`, 'PATCH', update);
      },

      regenerate(id) {
        return api.fetch(`v2/histograms/${id}/regenerate/`, 'POST');
      },

      findDocuments(id, path, value) {
        return api.fetch(`v2/histograms/${id}/find_documents/`, 'POST', {
          path,
          value
        });
      }

    },
    exports: {
      list(collectionId) {
        const query = {
          collection: collectionId
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/story-exports/?${qs}`);
      },

      get(id) {
        return api.fetch(`v2/story-exports/${id}/`);
      }

    },
    limits: {
      projects: {
        list(query = {}) {
          const qs = new URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/projects/?${qs}`);
        },

        activate(id) {
          return api.fetch(`v3/limits/projects/${id}/activate/`, 'POST');
        },

        deactivate(id) {
          return api.fetch(`v3/limits/projects/${id}/deactivate/`, 'POST');
        }

      },
      languages: {
        list(query = {}) {
          const qs = new URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/languages/?${qs}`);
        },

        activate(id) {
          return api.fetch(`v3/limits/languages/${id}/activate/`, 'POST');
        },

        deactivate(id) {
          return api.fetch(`v3/limits/languages/${id}/deactivate/`, 'POST');
        }

      },
      documents: {
        list(query = {}) {
          const qs = new URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/documents/?${qs}`);
        },

        statistics(query = {}) {
          const qs = new URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/documents/statistics/?${qs}`);
        }

      }
    }
  };
  api.fetch = fetch;
  return api;
};

/* eslint camelcase: "off" */

/* global URLSearchParams */
// TODO clean up old billing endpoints
const IDM = function (fetch) {
  const api = {
    billing: {
      createContact(group, data) {
        return api.fetch(`groups/${group}/contact/`, 'POST', data);
      },

      updateContact(group, contact, data) {
        return api.fetch(`groups/${group}/contact/${contact}/`, 'PUT', {
          data
        });
      },

      deleteContact(group, contact) {
        return api.fetch(`groups/${group}/contact/${contact}/`, 'DELETE');
      },

      getPaymentMethods(group) {
        return api.fetch(`groups/${group}/payment-method/`);
      },

      updatePaymentMethod(group, paymentMethod, data) {
        return api.fetch(`groups/${group}/payment-method/${paymentMethod}/`, 'PUT', {
          data
        });
      }

    },
    groups: {
      // TODO order
      list() {
        return api.fetch(`groups/`);
      },

      get(group) {
        return api.fetch(`groups/${group}/`);
      },

      chargeBeePortal(group) {
        return api.fetch(`groups/${group}/portal/`, 'POST');
      },

      checkout(group, payload) {
        return api.fetch(`groups/${group}/checkout/`, 'POST', payload);
      },

      estimateCheckout(payload) {
        return api.fetch(`groups/checkout-estimate/`, 'POST', payload);
      },

      prepareCheckout(payload) {
        return api.fetch(`groups/checkout-prepare/`, 'POST', payload);
      },

      invite(group, email, name, role) {
        return api.fetch(`groups/${group}/invitation/${email}/`, 'PUT', {
          name,
          role
        });
      },

      cancelInvite(group, email) {
        return api.fetch(`groups/${group}/invitation/${email}/`, 'DELETE');
      },

      removeMember(group, email) {
        return api.fetch(`groups/${group}/member/${email}/`, 'DELETE');
      },

      changeRole(group, email, role) {
        return api.fetch(`groups/${group}/role/${email}/`, 'POST', {
          role
        });
      },

      updatePrimaryPaymentMethod(group, payload) {
        return api.fetch(`groups/${group}/update-primary-payment-method/`, 'POST', payload);
      },

      scheduleLicenseChanges(group, plan) {
        return api.fetch(`groups/${group}/schedule-license-changes/`, 'POST', {
          plan
        });
      },

      cancelScheduledLicenseChanges(group) {
        return api.fetch(`groups/${group}/cancel-scheduled-license-changes/`, 'POST');
      },

      cancelLicense(group) {
        return api.fetch(`groups/${group}/cancel-license/`, 'POST');
      },

      getBillingInfo(group) {
        return api.fetch(`groups/${group}/billing-info/`);
      },

      updateBillingAddress(group, billing_address) {
        return api.fetch(`groups/${group}/billing-address/`, 'PUT', billing_address);
      }

    },
    users: {
      getMe() {
        return api.fetch(`user/`);
      },

      save(user) {
        return api.fetch('user/', 'PATCH', user);
      },

      getByEmail(email) {
        const qs = new URLSearchParams({
          email
        }).toString();
        return api.fetch(`user/?${qs}`);
      },

      saveSettings(settings) {
        return api.fetch('user/', 'PATCH', {
          settings
        });
      },

      acceptInvitation(user, group) {
        return api.fetch(`user/${user}/accept-invitation/${group}/`, 'POST');
      }

    },

    tokenExchange(refresh_token) {
      const headers = {
        'Content-Type': 'application/json'
      };
      return fetch('token-exchange/', 'POST', {
        refresh_token
      }, headers);
    }

  };
  api.fetch = fetch;
  return api;
};

/* global URLSearchParams */

const lexicon = function (fetch, baseUrl, token) {
  const api = {
    nouns: {
      list(trainings, options = {}) {
        const query = {
          trainings,
          page_size: options.page_size || options.pageSize,
          language: options.language,
          search: options.search,
          o: options.modified
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`nouns/?${qs}`);
      },

      create(lexiconEntry) {
        return api.fetch('nouns/', 'POST', lexiconEntry);
      },

      delete(id) {
        return api.fetch(`nouns/${id}/`, 'DELETE');
      },

      get(id) {
        return api.fetch(`nouns/${id}/`);
      },

      update(lexiconEntry) {
        return api.fetch(`nouns/${lexiconEntry.id}/`, 'PATCH', lexiconEntry);
      },

      lookupLemma(trainingId, language, lemma) {
        return api.fetch(`nouns/lemma-lookup/${trainingId}/${language}/${lemma}/`);
      }

    },
    verbs: {
      list(trainings, options = {}) {
        const query = {
          trainings,
          page_size: options.page_size || options.pageSize,
          language: options.language,
          search: options.search,
          o: options.modified
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`verbs/?${qs}`);
      },

      create(lexiconEntry) {
        return api.fetch('verbs/', 'POST', lexiconEntry);
      },

      delete(id) {
        return api.fetch(`verbs/${id}/`, 'DELETE');
      },

      addTraining(id, trainingId) {
        return api.fetch(`verbs/${id}/add-training/`, 'POST', {
          training: trainingId
        });
      },

      removeTraining(id, trainingId) {
        return api.fetch(`verbs/${id}/remove-training/`, 'POST', {
          training: trainingId
        });
      },

      get(id) {
        return api.fetch(`verbs/${id}/`);
      },

      update(lexiconEntry) {
        return api.fetch(`verbs/${lexiconEntry.id}/`, 'PATCH', lexiconEntry);
      },

      lookupLemma(trainingId, language, lemma) {
        return api.fetch(`verbs/lemma-lookup/${trainingId}/${language}/${lemma}/`);
      }

    },
    adjectives: {
      list(trainings, options = {}) {
        const query = {
          trainings,
          page_size: options.page_size || options.pageSize,
          language: options.language,
          search: options.search,
          o: options.modified
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`adjectives/?${qs}`);
      },

      create(lexiconEntry) {
        return api.fetch('adjectives/', 'POST', lexiconEntry);
      },

      delete(id) {
        return api.fetch(`adjectives/${id}/`, 'DELETE');
      },

      addTraining(id, trainingId) {
        return api.fetch(`adjectives/${id}/add-training/`, 'POST', {
          training: trainingId
        });
      },

      removeTraining(id, trainingId) {
        return api.fetch(`adjectives/${id}/remove-training/`, 'POST', {
          training: trainingId
        });
      },

      get(id) {
        return api.fetch(`adjectives/${id}/`);
      },

      update(lexiconEntry) {
        return api.fetch(`adjectives/${lexiconEntry.id}/`, 'PATCH', lexiconEntry);
      },

      lookupLemma(trainingId, language, lemma) {
        return api.fetch(`adjectives/lemma-lookup/${trainingId}/${language}/${lemma}/`);
      }

    },
    languageDefinitions: {
      list() {
        return api.fetch('definitions/');
      },

      get(language) {
        return api.fetch(`definitions/${language}/`);
      }

    }
  };
  api.fetch = fetch;
  return api;
};

/* eslint camelcase: "off" */

const BulkUpload = function (fetch, idToken) {
  const api = {
    uploads: {
      list() {
        return api.fetch(`uploads/?page_size=100`);
      },

      get(uploadId) {
        return api.fetch(`uploads/${uploadId}/`);
      },

      upload(collectionId, hint, file) {
        const data = new FormData();
        data.append('collection_id', collectionId);
        data.append('hint', hint);
        data.append('data_file', file);
        return api.fetch(`uploads/`, 'POST', data);
      },

      getFileLink(id) {
        return api.fetch(`uploads/${id}/`);
      }

    },
    itemResponses: {
      listByUpload(uploadId, options = {}) {
        const query = {
          upload: uploadId,
          only_user_errors: options.errorsOnly,
          page_size: 100
        };
        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`item-responses/?${qs}`);
      }

    }
  };

  api.fetch = function (url, method, body) {
    const headers = {
      authorization: `JWT ${idToken}`,
      Accept: 'application/json'
    };

    if (!(body instanceof FormData)) {
      headers['Content-Type'] = 'application/json';
    }

    return fetch(url, method, body, headers);
  };

  return api;
};

/* eslint camelcase: "off" */

/* global URLSearchParams */
const Quest = function (fetch) {
  const api = {
    quests: {
      list() {
        return api.fetch(`v1/quests/`);
      },

      get(id) {
        return api.fetch(`v1/quests/${id}/`);
      },

      finish(id, email) {
        return api.fetch(`v1/quests/${id}/finish/`, 'POST', {
          email
        });
      }

    },
    tasks: {
      list() {
        return api.fetch(`v1/tasks/`);
      },

      get(id) {
        return api.fetch(`v1/tasks/${id}/`);
      },

      finish(id, email) {
        return api.fetch(`v1/tasks/${id}/finish/`, 'POST', {
          email
        });
      }

    },
    users: {
      get(email) {
        const qs = new URLSearchParams({
          email
        }).toString();
        return api.fetch(`v1/users/?${qs}`);
      }

    },
    lessons: {
      list() {
        return api.fetch(`v1/lessons/`);
      },

      get(id) {
        return api.fetch(`v1/lessons/${id}/`);
      },

      create(lesson) {
        return api.fetch(`v1/lessons/`, 'POST', lesson);
      },

      // can only update linked_project and last_valid_route
      update(id, update) {
        return api.fetch(`v1/lessons/${id}/`, 'PATCH', update);
      },

      addStep(id, stepName) {
        return api.fetch(`v1/lessons/${id}/steps/`, 'POST', {
          step: stepName
        });
      },

      finish(id) {
        return api.fetch(`v1/lessons/${id}/finish/`, 'POST');
      }

    }
  };
  api.fetch = fetch;
  return api;
};

/* eslint camelcase: "off" */

/* global URLSearchParams */

const Report = function (fetch, idToken) {
  const api = {
    reports: {
      list(licenseGroup, filters = {}, options = {}) {
        const query = Object.assign({}, filters, {
          license_group: licenseGroup,
          page: options.page,
          page_size: options.pageSize,
          collection: undefined,
          project: undefined
        });

        if (filters.collection) {
          query.scope = 'collection';
          query.subject = filters.collection;
        } else if (filters.project) {
          query.scope = 'project';
          query.subject = filters.project;
        }

        const qs = new URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v3/reports/?${qs}`);
      },

      get(id) {
        return api.fetch(`v3/reports/${id}/`);
      }

    }
  };

  api.fetch = function (url, method) {
    // *allow* caching
    const headers = {
      Authorization: `JWT ${idToken}`,
      'Content-Type': 'application/json'
    };
    return fetch(url, method, undefined, headers);
  };

  return api;
};

/* global FormData, fetch */
//
// The client operates on plain js objects and quasi-static procedures.
// It also unifies both the myax and training apis into one interface.

class AxSemanticsClient extends EventEmitter {
  constructor(userConfig) {
    super();
    const config = {
      myAxBaseUrl: 'https://api.ax-semantics.com/',
      trainingBaseUrl: 'https://training-api.ax-semantics.com/',
      lexiconBaseUrl: 'https://lexicon.ax-semantics.com/v1/',
      idmBaseUrl: 'https://idm.ax-semantics.com/v1/',
      bulkUploadBaseUrl: 'https://bulk-api.ax-semantics.com/v1/',
      questBaseUrl: 'https://quest-api.ax-semantics.com/',
      reportBaseUrl: 'https://report-api.ax-semantics.com/',
      idToken: '',
      refreshToken: null,
      fetch: AxSemanticsClient.fetch // set this for handling the fetch promise globally

    };
    Object.assign(config, userConfig);
    this._idm = IDM(config.fetch.bind(this, config.idmBaseUrl)); // needed for refresh token shake

    const initApis = idToken => {
      this.idToken = idToken;
      this._myax = MyAx(config.fetch.bind(this, config.myAxBaseUrl));
      this._editor = Training(config.fetch.bind(this, config.trainingBaseUrl));
      this._lexicon = lexicon(config.fetch.bind(this, config.lexiconBaseUrl));
      this._bulkUpload = BulkUpload(config.fetch.bind(this, config.bulkUploadBaseUrl), this.idToken);
      this._quest = Quest(config.fetch.bind(this, config.questBaseUrl));
      this._report = Report(config.fetch.bind(this, config.reportBaseUrl), this.idToken);
    };

    if (config.refreshToken) {
      this._idm.tokenExchange(config.refreshToken).then(tokenExchange => {
        initApis(tokenExchange.id_token);
        this.emit('ready');
      }).catch(error => {
        this.emit('error', error);
      });
    } else {
      initApis(config.idToken);
      this.emit('ready');
    }
  }

  static fetch(baseUrl, url, method, body, userHeaders) {
    const headers = userHeaders || {
      'Content-Type': 'application/json',
      authorization: `JWT ${this.idToken}`,
      'Cache-Control': 'no-cache'
    };
    const options = {
      method: method || 'GET',
      headers,
      body: body instanceof FormData ? body : JSON.stringify(body),
      credentials: 'omit' // api wants no cookies

    };
    return fetch(url.startsWith('http') ? url : baseUrl + url, options).then(response => {
      if (response.status === 204 || response.ok && response.headers.get('Content-Type') !== 'application/json') return Promise.resolve(); // no content to parse

      return response.json().then(json => {
        if (!response.ok) // TODO change to error
          return Promise.reject({
            response,
            json
          }); // eslint-disable-line

        return Promise.resolve(json);
      }).catch(error => {
        if (!error.json) // no json
          return Promise.reject({
            response,
            error
          }); // eslint-disable-line

        return Promise.reject(error);
      });
    }).catch(error => {
      return Promise.reject(error);
    });
  }

}

module.exports = AxSemanticsClient;
