'use strict';

var FormData = require('form-data');
var fetch = require('node-fetch');
var EventEmitter = require('events');
var url = require('url');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var FormData__default = /*#__PURE__*/_interopDefaultLegacy(FormData);
var fetch__default = /*#__PURE__*/_interopDefaultLegacy(fetch);
var EventEmitter__default = /*#__PURE__*/_interopDefaultLegacy(EventEmitter);

function cleanQuery(object) {
  Object.keys(object).forEach(key => !object[key] && delete object[key]);
  return object;
}
function cleanNullsorUndefined(object) {
  Object.keys(object).forEach(key => (object[key] === null || object[key] === undefined) && delete object[key]);
  return object;
}

const Training = function (fetch, baseUrl, token) {
  const api = {
    activityLogs: {
      list(trainingId, filters = {}, options = {}) {
        const query = {
          training: trainingId,
          page: options.page
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v1/activity-logs/?${qs}`);
      },

      get(id) {
        return api.fetch(`v1/activity-logs/${id}/`);
      }

    },
    languages: {
      list() {
        return api.fetch(`v1/languages/`);
      },

      listAll() {
        return api.fetch(`v1/languages/all/`);
      }

    },
    translationPackages: {
      list({
        projectId,
        assignee,
        state
      } = {}) {
        const query = {
          project: projectId,
          assignee,
          state
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v3/translation-packages/?${qs}`);
      },

      get(id) {
        return api.fetch(`v3/translation-packages/${id}/`);
      },

      create(translationPackage) {
        return api.fetch(`v3/translation-packages/`, 'POST', translationPackage);
      },

      delete(id) {
        return api.fetch(`v3/translation-packages/${id}/`, 'DELETE');
      },

      update(translationPackageId, update) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/`, 'PATCH', update);
      },

      simulate(packageOptions) {
        return api.fetch('v3/translation-packages/simulate/', 'POST', packageOptions);
      }

    },
    translationItems: {
      list(translationPackageId) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/`);
      },

      create({
        translationPackageId,
        translationItem
      }) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/`, 'POST', translationItem);
      },

      delete({
        translationPackageId,
        objectType,
        objectId
      }) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/${objectType}-${objectId}/`, 'DELETE');
      },

      update({
        translationPackageId,
        objectType,
        objectId,
        ...update
      }) {
        return api.fetch(`v3/translation-packages/${translationPackageId}/translation-items/${objectType}-${objectId}/`, 'PATCH', update);
      }

    },
    notifications: {
      list() {
        return api.fetch(`v3/notifications/`);
      },

      get(id) {
        return api.fetch(`v3/notifications/${id}/`);
      },

      update(id, patchset) {
        return api.fetch(`v3/notifications/${id}/`, 'PATCH', patchset);
      },

      getUnreadCount() {
        return api.fetch('v3/notifications/unread-count/');
      },

      markAllAsRead() {
        return api.fetch('v3/notifications/mark-all-as-read/', 'POST');
      }

    },
    trainings: {
      list(fields = 'id,name') {
        return api.fetch(`v1/trainings/?fields=${fields}`);
      },

      get(id) {
        return api.fetch(`v1/trainings/${id}/`);
      },

      properties(id) {
        return api.fetch(`v1/properties/?training=${id}`);
      },

      create(name, license_holder, more) {
        return api.fetch(`v1/trainings/`, 'POST', {
          name,
          license_holder,
          ...more
        });
      },

      update(id, patchset) {
        return api.fetch(`v1/trainings/${id}/`, 'PATCH', patchset);
      },

      delete(id) {
        return api.fetch(`v1/trainings/${id}/`, 'DELETE');
      },

      publish(id) {
        return api.fetch(`v1/trainings/${id}/publish/`, 'POST');
      }

    },
    // https://github.com/aexeagmbh/rincewind/issues/317#issuecomment-286102626
    validationData: {
      list(trainingId) {
        return api.fetch(`v1/validation-data/?training=${trainingId}`);
      },

      get(id) {
        return api.fetch(`v1/validation-data/${id}/`);
      },

      create(validationData) {
        // {training, name, uid, data}
        return api.fetch(`v1/validation-data/`, 'POST', validationData);
      },

      createSyncedDocument(trainingId, collection_id, document) {
        return api.fetch(`v1/validation-data/create-synced/`, 'POST', {
          training: trainingId,
          name: document.name,
          uid: document.uid,
          data: document.blob,
          myax_api_version: 2,
          myax_reference: {
            collection_id,
            document_id: document.id
          }
        });
      },

      update(validationData) {
        const {
          id,
          name,
          uid,
          data
        } = validationData;
        return api.fetch(`v1/validation-data/${id}/`, 'PATCH', {
          name,
          uid,
          data
        });
      },

      delete(id) {
        return api.fetch(`v1/validation-data/${id}/`, 'DELETE');
      },

      select(id, language) {
        return api.fetch(`v1/validation-data/${id}/select/`, 'PUT', {
          language
        });
      },

      selectDefault(id, language) {
        return api.fetch(`v1/validation-data/${id}/select-global/`, 'PUT', {
          language
        });
      },

      sync(validationData) {
        const {
          id,
          name,
          uid,
          data
        } = validationData;
        return api.fetch(`v1/validation-data/${id}/sync/`, 'PUT', {
          name,
          uid,
          data
        });
      }

    },
    import: {
      import(trainingId, atml) {
        return api.fetch(`v1/trainings/${trainingId}/import/`, 'PUT', {
          atml
        });
      },

      migrate(oldTrainingId, newTrainingId, options) {
        return api.fetch(`v1/trainings/${newTrainingId}/migrate-from-rincewind/`, 'POST', {
          rincewind_training_id: oldTrainingId,
          copy_collections: options && options.copyCollections,
          parametrize_expressions: options && options.parametrizeExpressions
        });
      },

      status(trainingId) {
        return api.fetch(`v1/trainings/${trainingId}/import/?latest=1`);
      }

    }
  };
  api.fetch = fetch;
  return api;
};

const fieldsFromOptions = function (options) {
  return options.fields instanceof Array ? options.fields.join(',') : options.fields;
};

const MyAx = function (fetch, baseUrl, token) {
  const api = {
    instantGenerationEndpoints: {
      list(training_id, options = {}) {
        const query = {
          training_id,
          page_size: options.pageSize || 50,
          fields: fieldsFromOptions(options),
          is_instant: true
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/collections/?${qs}`);
      },

      get(id, options = {}) {
        return api.collections.get(id, options);
      },

      create(endpoint) {
        endpoint.is_instant = true;
        return api.fetch(`v2/collections/`, 'POST', endpoint);
      }

    },
    collections: {
      // options = {page_size: 50, fields}
      list(training_id, options = {}) {
        const query = {
          training_id,
          page_size: options.pageSize || 50,
          fields: fieldsFromOptions(options),
          is_instant: false
        };
        const qs = new url.URLSearchParams(cleanNullsorUndefined(query)).toString();
        return api.fetch(`v2/collections/?${qs}`);
      },

      get(id, options = {}) {
        const query = {
          fields: fieldsFromOptions(options)
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/collections/${id}/?${qs}`);
      },

      getProcessingStates(collectionId) {
        return api.fetch(`v2/collections/${collectionId}/document-processing-states/`);
      },

      // collection = { language, training_id, name, licence_holder, training_tag: 'draft' }
      create(collection) {
        return api.fetch(`v2/collections/`, 'POST', collection);
      },

      delete(id) {
        return api.fetch(`v2/collections/${id}/`, 'DELETE');
      },

      save(collection) {
        return api.fetch(`v2/collections/${collection.id}/`, 'PATCH', {
          name: collection.name,
          language: collection.language,
          uses_published_training: collection.uses_published_training,
          webhook_secret: collection.webhook_secret,
          webhook_url: collection.webhook_url,
          fail_webhook_url: collection.fail_webhook_url,
          flags: collection.flags
        });
      },

      generateAll(id) {
        return api.fetch(`v2/collections/${id}/generate-content/`, 'POST');
      },

      generateFiltered(id, search, processingState) {
        const query = {
          search,
          processing_state: processingState
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/collections/${id}/generate-content/?${qs}`, 'POST');
      },

      empty(id) {
        return api.fetch(`v2/collections/${id}/empty/`, 'POST');
      }

    },
    // https://github.com/aexeagmbh/myax/blob/master/myax/document/processing_states.py
    // optional search and processingState ['none', 'requested', 'generated', 'delivered', 'failed']
    documents: {
      // filters = {search, processingState}
      // options = {fields}
      list(collectionId, filters = {}, options = {}) {
        const query = {
          collection: collectionId,
          search: filters.search,
          processing_state: filters.processingState,
          fields: options.fields
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/documents/?${qs}`);
      },

      // blob has to contain 'name' and 'uid'
      create(collectionId, blob) {
        return api.fetch(`v2/collections/${collectionId}/document/`, 'POST', blob);
      },

      // blob has to contain 'name'
      updateOrCreateByUid(collectionId, blob, uid) {
        const payload = Object.assign({
          uid
        }, blob);
        return api.fetch(`v2/collections/${collectionId}/update-or-create-by-uid/?uid=${uid}`, 'POST', payload);
      },

      // options = {fields}
      get(id, options = {}) {
        const query = {
          fields: fieldsFromOptions(options)
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/documents/${id}/?${qs}`);
      },

      generate(id) {
        return api.fetch(`v2/documents/${id}/generate-content/?highprio`, 'POST');
      },

      delete(documentId) {
        return api.fetch(`v2/documents/${documentId}/`, 'DELETE');
      }

    },
    histograms: {
      list(collectionOrProjectId, filters = {
        pageSize: 100
      }) {
        const query = {};

        if (typeof collectionOrProjectId === 'number' || /^[0-9]+$/.test(collectionOrProjectId)) {
          query.collection = collectionOrProjectId;
        } else if (collectionOrProjectId) {
          query.project = collectionOrProjectId;
        }

        if (filters && filters.latest) query.latest = true;
        if (filters && filters.pageSize) query.page_size = filters.pageSize;
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/histograms/?${qs}`);
      },

      create(collectionIdOrData) {
        let data;

        if (typeof collectionIdOrData === 'object') {
          data = collectionIdOrData;
        } else {
          data = {
            collection: collectionIdOrData
          };
        }

        return api.fetch(`v2/histograms/`, 'POST', data);
      },

      delete(id) {
        return api.fetch(`v2/histograms/${id}/`, 'DELETE');
      },

      get(id) {
        return api.fetch(`v2/histograms/${id}/`);
      },

      update(id, update) {
        return api.fetch(`v2/histograms/${id}/`, 'PATCH', update);
      },

      regenerate(id) {
        return api.fetch(`v2/histograms/${id}/regenerate/`, 'POST');
      },

      findDocuments(id, path, value) {
        return api.fetch(`v2/histograms/${id}/find_documents/`, 'POST', {
          path,
          value
        });
      }

    },
    exports: {
      list(collectionId) {
        const query = {
          collection: collectionId
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v2/story-exports/?${qs}`);
      },

      get(id) {
        return api.fetch(`v2/story-exports/${id}/`);
      }

    },
    limits: {
      projects: {
        list(query = {}) {
          const qs = new url.URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/projects/?${qs}`);
        },

        activate(id) {
          return api.fetch(`v3/limits/projects/${id}/activate/`, 'POST');
        },

        deactivate(id) {
          return api.fetch(`v3/limits/projects/${id}/deactivate/`, 'POST');
        }

      },
      languages: {
        list(query = {}) {
          const qs = new url.URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/languages/?${qs}`);
        },

        activate(id) {
          return api.fetch(`v3/limits/languages/${id}/activate/`, 'POST');
        },

        deactivate(id) {
          return api.fetch(`v3/limits/languages/${id}/deactivate/`, 'POST');
        }

      },
      documents: {
        list(query = {}) {
          const qs = new url.URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/documents/?${qs}`);
        },

        statistics(query = {}) {
          const qs = new url.URLSearchParams(cleanQuery(query)).toString();
          return api.fetch(`v3/limits/documents/statistics/?${qs}`);
        }

      }
    }
  };
  api.fetch = fetch;
  return api;
};

/* eslint camelcase: "off" */

/* global URLSearchParams */
// TODO clean up old billing endpoints
const IDM = function (fetch) {
  const api = {
    billing: {
      createContact(group, data) {
        return api.fetch(`groups/${group}/contact/`, 'POST', data);
      },

      updateContact(group, contact, data) {
        return api.fetch(`groups/${group}/contact/${contact}/`, 'PUT', {
          data
        });
      },

      deleteContact(group, contact) {
        return api.fetch(`groups/${group}/contact/${contact}/`, 'DELETE');
      },

      getPaymentMethods(group) {
        return api.fetch(`groups/${group}/payment-method/`);
      },

      updatePaymentMethod(group, paymentMethod, data) {
        return api.fetch(`groups/${group}/payment-method/${paymentMethod}/`, 'PUT', {
          data
        });
      }

    },
    groups: {
      // TODO order
      list() {
        return api.fetch(`groups/`);
      },

      get(group) {
        return api.fetch(`groups/${group}/`);
      },

      chargeBeePortal(group) {
        return api.fetch(`groups/${group}/portal/`, 'POST');
      },

      checkout(group, payload) {
        return api.fetch(`groups/${group}/checkout/`, 'POST', payload);
      },

      estimateCheckout(payload) {
        return api.fetch(`groups/checkout-estimate/`, 'POST', payload);
      },

      prepareCheckout(payload) {
        return api.fetch(`groups/checkout-prepare/`, 'POST', payload);
      },

      invite(group, email, name, role) {
        return api.fetch(`groups/${group}/invitation/${email}/`, 'PUT', {
          name,
          role
        });
      },

      cancelInvite(group, email) {
        return api.fetch(`groups/${group}/invitation/${email}/`, 'DELETE');
      },

      removeMember(group, email) {
        return api.fetch(`groups/${group}/member/${email}/`, 'DELETE');
      },

      changeRole(group, email, role) {
        return api.fetch(`groups/${group}/role/${email}/`, 'POST', {
          role
        });
      },

      updatePrimaryPaymentMethod(group, payload) {
        return api.fetch(`groups/${group}/update-primary-payment-method/`, 'POST', payload);
      },

      scheduleLicenseChanges(group, plan) {
        return api.fetch(`groups/${group}/schedule-license-changes/`, 'POST', {
          plan
        });
      },

      cancelScheduledLicenseChanges(group) {
        return api.fetch(`groups/${group}/cancel-scheduled-license-changes/`, 'POST');
      },

      cancelLicense(group) {
        return api.fetch(`groups/${group}/cancel-license/`, 'POST');
      },

      getBillingInfo(group) {
        return api.fetch(`groups/${group}/billing-info/`);
      },

      updateBillingAddress(group, billing_address) {
        return api.fetch(`groups/${group}/billing-address/`, 'PUT', billing_address);
      }

    },
    users: {
      getMe() {
        return api.fetch(`user/`);
      },

      save(user) {
        return api.fetch('user/', 'PATCH', user);
      },

      getByEmail(email) {
        const qs = new url.URLSearchParams({
          email
        }).toString();
        return api.fetch(`user/?${qs}`);
      },

      saveSettings(settings) {
        return api.fetch('user/', 'PATCH', {
          settings
        });
      },

      acceptInvitation(user, group) {
        return api.fetch(`user/${user}/accept-invitation/${group}/`, 'POST');
      }

    },

    tokenExchange(refresh_token) {
      const headers = {
        'Content-Type': 'application/json'
      };
      return fetch('token-exchange/', 'POST', {
        refresh_token
      }, headers);
    }

  };
  api.fetch = fetch;
  return api;
};

const lexicon = function (fetch, baseUrl, token) {
  const api = {
    nouns: {
      list(trainings, options = {}) {
        const query = {
          trainings,
          page_size: options.page_size || options.pageSize,
          language: options.language,
          search: options.search,
          o: options.modified
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`nouns/?${qs}`);
      },

      create(lexiconEntry) {
        return api.fetch('nouns/', 'POST', lexiconEntry);
      },

      delete(id) {
        return api.fetch(`nouns/${id}/`, 'DELETE');
      },

      get(id) {
        return api.fetch(`nouns/${id}/`);
      },

      update(lexiconEntry) {
        return api.fetch(`nouns/${lexiconEntry.id}/`, 'PATCH', lexiconEntry);
      },

      lookupLemma(trainingId, language, lemma) {
        return api.fetch(`nouns/lemma-lookup/${trainingId}/${language}/${lemma}/`);
      }

    },
    verbs: {
      list(trainings, options = {}) {
        const query = {
          trainings,
          page_size: options.page_size || options.pageSize,
          language: options.language,
          search: options.search,
          o: options.modified
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`verbs/?${qs}`);
      },

      create(lexiconEntry) {
        return api.fetch('verbs/', 'POST', lexiconEntry);
      },

      delete(id) {
        return api.fetch(`verbs/${id}/`, 'DELETE');
      },

      addTraining(id, trainingId) {
        return api.fetch(`verbs/${id}/add-training/`, 'POST', {
          training: trainingId
        });
      },

      removeTraining(id, trainingId) {
        return api.fetch(`verbs/${id}/remove-training/`, 'POST', {
          training: trainingId
        });
      },

      get(id) {
        return api.fetch(`verbs/${id}/`);
      },

      update(lexiconEntry) {
        return api.fetch(`verbs/${lexiconEntry.id}/`, 'PATCH', lexiconEntry);
      },

      lookupLemma(trainingId, language, lemma) {
        return api.fetch(`verbs/lemma-lookup/${trainingId}/${language}/${lemma}/`);
      }

    },
    adjectives: {
      list(trainings, options = {}) {
        const query = {
          trainings,
          page_size: options.page_size || options.pageSize,
          language: options.language,
          search: options.search,
          o: options.modified
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`adjectives/?${qs}`);
      },

      create(lexiconEntry) {
        return api.fetch('adjectives/', 'POST', lexiconEntry);
      },

      delete(id) {
        return api.fetch(`adjectives/${id}/`, 'DELETE');
      },

      addTraining(id, trainingId) {
        return api.fetch(`adjectives/${id}/add-training/`, 'POST', {
          training: trainingId
        });
      },

      removeTraining(id, trainingId) {
        return api.fetch(`adjectives/${id}/remove-training/`, 'POST', {
          training: trainingId
        });
      },

      get(id) {
        return api.fetch(`adjectives/${id}/`);
      },

      update(lexiconEntry) {
        return api.fetch(`adjectives/${lexiconEntry.id}/`, 'PATCH', lexiconEntry);
      },

      lookupLemma(trainingId, language, lemma) {
        return api.fetch(`adjectives/lemma-lookup/${trainingId}/${language}/${lemma}/`);
      }

    },
    languageDefinitions: {
      list() {
        return api.fetch('definitions/');
      },

      get(language) {
        return api.fetch(`definitions/${language}/`);
      }

    }
  };
  api.fetch = fetch;
  return api;
};

const BulkUpload = function (fetch, idToken) {
  const api = {
    uploads: {
      list() {
        return api.fetch(`uploads/?page_size=100`);
      },

      get(uploadId) {
        return api.fetch(`uploads/${uploadId}/`);
      },

      upload(collectionId, hint, file) {
        const data = new FormData__default["default"]();
        data.append('collection_id', collectionId);
        data.append('hint', hint);
        data.append('data_file', file);
        return api.fetch(`uploads/`, 'POST', data);
      },

      getFileLink(id) {
        return api.fetch(`uploads/${id}/`);
      }

    },
    itemResponses: {
      listByUpload(uploadId, options = {}) {
        const query = {
          upload: uploadId,
          only_user_errors: options.errorsOnly,
          page_size: 100
        };
        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`item-responses/?${qs}`);
      }

    }
  };

  api.fetch = function (url, method, body) {
    const headers = {
      authorization: `JWT ${idToken}`,
      Accept: 'application/json'
    };

    if (!(body instanceof FormData__default["default"])) {
      headers['Content-Type'] = 'application/json';
    }

    return fetch(url, method, body, headers);
  };

  return api;
};

/* eslint camelcase: "off" */

/* global URLSearchParams */
const Quest = function (fetch) {
  const api = {
    quests: {
      list() {
        return api.fetch(`v1/quests/`);
      },

      get(id) {
        return api.fetch(`v1/quests/${id}/`);
      },

      finish(id, email) {
        return api.fetch(`v1/quests/${id}/finish/`, 'POST', {
          email
        });
      }

    },
    tasks: {
      list() {
        return api.fetch(`v1/tasks/`);
      },

      get(id) {
        return api.fetch(`v1/tasks/${id}/`);
      },

      finish(id, email) {
        return api.fetch(`v1/tasks/${id}/finish/`, 'POST', {
          email
        });
      }

    },
    users: {
      get(email) {
        const qs = new url.URLSearchParams({
          email
        }).toString();
        return api.fetch(`v1/users/?${qs}`);
      }

    },
    lessons: {
      list() {
        return api.fetch(`v1/lessons/`);
      },

      get(id) {
        return api.fetch(`v1/lessons/${id}/`);
      },

      create(lesson) {
        return api.fetch(`v1/lessons/`, 'POST', lesson);
      },

      // can only update linked_project and last_valid_route
      update(id, update) {
        return api.fetch(`v1/lessons/${id}/`, 'PATCH', update);
      },

      addStep(id, stepName) {
        return api.fetch(`v1/lessons/${id}/steps/`, 'POST', {
          step: stepName
        });
      },

      finish(id) {
        return api.fetch(`v1/lessons/${id}/finish/`, 'POST');
      }

    }
  };
  api.fetch = fetch;
  return api;
};

/* eslint camelcase: "off" */

/* global URLSearchParams */

const Report = function (fetch, idToken) {
  const api = {
    reports: {
      list(licenseGroup, filters = {}, options = {}) {
        const query = Object.assign({}, filters, {
          license_group: licenseGroup,
          page: options.page,
          page_size: options.pageSize,
          collection: undefined,
          project: undefined
        });

        if (filters.collection) {
          query.scope = 'collection';
          query.subject = filters.collection;
        } else if (filters.project) {
          query.scope = 'project';
          query.subject = filters.project;
        }

        const qs = new url.URLSearchParams(cleanQuery(query)).toString();
        return api.fetch(`v3/reports/?${qs}`);
      },

      get(id) {
        return api.fetch(`v3/reports/${id}/`);
      }

    }
  };

  api.fetch = function (url, method) {
    // *allow* caching
    const headers = {
      Authorization: `JWT ${idToken}`,
      'Content-Type': 'application/json'
    };
    return fetch(url, method, undefined, headers);
  };

  return api;
};

//
// The client operates on plain js objects and quasi-static procedures.
// It also unifies both the myax and training apis into one interface.

class AxSemanticsClient extends EventEmitter__default["default"] {
  constructor(userConfig) {
    super();
    const config = {
      myAxBaseUrl: 'https://api.ax-semantics.com/',
      trainingBaseUrl: 'https://training-api.ax-semantics.com/',
      lexiconBaseUrl: 'https://lexicon.ax-semantics.com/v1/',
      idmBaseUrl: 'https://idm.ax-semantics.com/v1/',
      bulkUploadBaseUrl: 'https://bulk-api.ax-semantics.com/v1/',
      questBaseUrl: 'https://quest-api.ax-semantics.com/',
      reportBaseUrl: 'https://report-api.ax-semantics.com/',
      idToken: '',
      refreshToken: null,
      fetch: AxSemanticsClient.fetch // set this for handling the fetch promise globally

    };
    Object.assign(config, userConfig);
    this._idm = IDM(config.fetch.bind(this, config.idmBaseUrl)); // needed for refresh token shake

    const initApis = idToken => {
      this.idToken = idToken;
      this._myax = MyAx(config.fetch.bind(this, config.myAxBaseUrl));
      this._editor = Training(config.fetch.bind(this, config.trainingBaseUrl));
      this._lexicon = lexicon(config.fetch.bind(this, config.lexiconBaseUrl));
      this._bulkUpload = BulkUpload(config.fetch.bind(this, config.bulkUploadBaseUrl), this.idToken);
      this._quest = Quest(config.fetch.bind(this, config.questBaseUrl));
      this._report = Report(config.fetch.bind(this, config.reportBaseUrl), this.idToken);
    };

    if (config.refreshToken) {
      this._idm.tokenExchange(config.refreshToken).then(tokenExchange => {
        initApis(tokenExchange.id_token);
        this.emit('ready');
      }).catch(error => {
        this.emit('error', error);
      });
    } else {
      initApis(config.idToken);
      this.emit('ready');
    }
  }

  static fetch(baseUrl, url, method, body, userHeaders) {
    const headers = userHeaders || {
      'Content-Type': 'application/json',
      authorization: `JWT ${this.idToken}`,
      'Cache-Control': 'no-cache'
    };
    const options = {
      method: method || 'GET',
      headers,
      body: body instanceof FormData__default["default"] ? body : JSON.stringify(body),
      credentials: 'omit' // api wants no cookies

    };
    return fetch__default["default"](url.startsWith('http') ? url : baseUrl + url, options).then(response => {
      if (response.status === 204 || response.ok && response.headers.get('Content-Type') !== 'application/json') return Promise.resolve(); // no content to parse

      return response.json().then(json => {
        if (!response.ok) // TODO change to error
          return Promise.reject({
            response,
            json
          }); // eslint-disable-line

        return Promise.resolve(json);
      }).catch(error => {
        if (!error.json) // no json
          return Promise.reject({
            response,
            error
          }); // eslint-disable-line

        return Promise.reject(error);
      });
    }).catch(error => {
      return Promise.reject(error);
    });
  }

}

module.exports = AxSemanticsClient;
